@@ -5,6 +5,7 @@ import threading
 import unittest
 
 import fastrlock
+import pytest
 
 import cupy.cuda
 from cupy.cuda import device
@@ -39,16 +40,21 @@ class TestUnownedMemoryClass(unittest.TestCase):
 
 
 @testing.parameterize(*testing.product({
-    'allocator': [memory._malloc, memory.malloc_managed],
+    'allocator': [memory._malloc, memory.malloc_managed, memory.malloc_async],
     'specify_device_id': [True, False],
 }))
 @testing.gpu
 class TestUnownedMemory(unittest.TestCase):
 
     def check(self, device_id):
-        if (cupy.cuda.runtime.is_hip
-                and self.allocator is memory.malloc_managed):
-            raise unittest.SkipTest('HIP does not support managed memory')
+        if cupy.cuda.runtime.is_hip:
+            if self.allocator is memory.malloc_managed:
+                raise unittest.SkipTest('HIP does not support managed memory')
+            if self.allocator is memory.malloc_async:
+                raise unittest.SkipTest('HIP does not support async mempool')
+        elif cupy.cuda.driver.get_build_version() < 11020:
+            raise unittest.SkipTest('malloc_async is supported since '
+                                    'CUDA 11.2')
 
         size = 24
         shape = (2, 3)
@@ -124,6 +130,7 @@ class TestMemoryPointer(unittest.TestCase):
         a_gpu = memory.alloc(4)
         a_cpu = ctypes.c_int(100)
         a_gpu.copy_from(ctypes.cast(ctypes.byref(a_cpu), ctypes.c_void_p), 4)
+
         b_cpu = ctypes.c_int()
         a_gpu.copy_to_host(
             ctypes.cast(ctypes.byref(b_cpu), ctypes.c_void_p), 4)
@@ -141,6 +148,30 @@ class TestMemoryPointer(unittest.TestCase):
             ctypes.cast(ctypes.byref(b_cpu), ctypes.c_void_p), 4)
         assert b_cpu.value == a_cpu.value
 
+    def test_copy_to_and_from_host_using_raw_ptr(self):
+        a_gpu = memory.alloc(4)
+        a_cpu = ctypes.c_int(100)
+        a_cpu_ptr = ctypes.cast(ctypes.byref(a_cpu), ctypes.c_void_p)
+        a_gpu.copy_from(a_cpu_ptr.value, 4)
+
+        b_cpu = ctypes.c_int()
+        b_cpu_ptr = ctypes.cast(ctypes.byref(b_cpu), ctypes.c_void_p)
+        a_gpu.copy_to_host(b_cpu_ptr.value, 4)
+        assert b_cpu.value == a_cpu.value
+
+    def test_copy_from_device_using_raw_ptr(self):
+        a_gpu = memory.alloc(4)
+        a_cpu = ctypes.c_int(100)
+        a_cpu_ptr = ctypes.cast(ctypes.byref(a_cpu), ctypes.c_void_p)
+        a_gpu.copy_from(a_cpu_ptr.value, 4)
+
+        b_gpu = memory.alloc(4)
+        b_gpu.copy_from(a_gpu, 4)
+        b_cpu = ctypes.c_int()
+        b_cpu_ptr = ctypes.cast(ctypes.byref(b_cpu), ctypes.c_void_p)
+        b_gpu.copy_to_host(b_cpu_ptr.value, 4)
+        assert b_cpu.value == a_cpu.value
+
     def test_memset(self):
         a_gpu = memory.alloc(4)
         a_gpu.memset(1, 4)
@@ -152,6 +183,82 @@ class TestMemoryPointer(unittest.TestCase):
             a_gpu += 1
 
 
+@testing.parameterize(*testing.product({
+    'use_streams': [True, False],
+}))
+@testing.gpu
+class TestMemoryPointerAsync(unittest.TestCase):
+
+    def setUp(self):
+        self.stream = stream_module.Stream() if self.use_streams else None
+
+    def test_copy_to_and_from_host_async(self):
+        a_gpu = memory.alloc(4)
+        a_cpu = ctypes.c_int(100)
+        a_gpu.copy_from_async(ctypes.cast(ctypes.byref(
+            a_cpu), ctypes.c_void_p), 4, stream=self.stream)
+
+        b_cpu = ctypes.c_int()
+        a_gpu.copy_to_host_async(
+            ctypes.cast(ctypes.byref(b_cpu), ctypes.c_void_p),
+            4, stream=self.stream)
+        if self.stream is not None:
+            self.stream.synchronize()
+        else:
+            stream_module.get_current_stream().synchronize()
+        assert b_cpu.value == a_cpu.value
+
+    def test_copy_from_device_async(self):
+        a_gpu = memory.alloc(4)
+        a_cpu = ctypes.c_int(100)
+        a_gpu.copy_from_async(ctypes.cast(ctypes.byref(
+            a_cpu), ctypes.c_void_p), 4, stream=self.stream)
+
+        b_gpu = memory.alloc(4)
+        b_gpu.copy_from_async(a_gpu, 4, stream=self.stream)
+        b_cpu = ctypes.c_int()
+        b_gpu.copy_to_host_async(
+            ctypes.cast(ctypes.byref(b_cpu), ctypes.c_void_p),
+            4, stream=self.stream)
+        if self.stream is not None:
+            self.stream.synchronize()
+        else:
+            stream_module.get_current_stream().synchronize()
+        assert b_cpu.value == a_cpu.value
+
+    def test_copy_to_and_from_host_async_using_raw_ptr(self):
+        a_gpu = memory.alloc(4)
+        a_cpu = ctypes.c_int(100)
+        a_cpu_ptr = ctypes.cast(ctypes.byref(a_cpu), ctypes.c_void_p)
+        a_gpu.copy_from_async(a_cpu_ptr.value, 4, stream=self.stream)
+
+        b_cpu = ctypes.c_int()
+        b_cpu_ptr = ctypes.cast(ctypes.byref(b_cpu), ctypes.c_void_p)
+        a_gpu.copy_to_host_async(b_cpu_ptr.value, 4, stream=self.stream)
+        if self.stream is not None:
+            self.stream.synchronize()
+        else:
+            stream_module.get_current_stream().synchronize()
+        assert b_cpu.value == a_cpu.value
+
+    def test_copy_from_device_async_using_raw_ptr(self):
+        a_gpu = memory.alloc(4)
+        a_cpu = ctypes.c_int(100)
+        a_cpu_ptr = ctypes.cast(ctypes.byref(a_cpu), ctypes.c_void_p)
+        a_gpu.copy_from_async(a_cpu_ptr.value, 4, stream=self.stream)
+
+        b_gpu = memory.alloc(4)
+        b_gpu.copy_from_async(a_gpu, 4, stream=self.stream)
+        b_cpu = ctypes.c_int()
+        b_cpu_ptr = ctypes.cast(ctypes.byref(b_cpu), ctypes.c_void_p)
+        b_gpu.copy_to_host_async(b_cpu_ptr.value, 4, stream=self.stream)
+        if self.stream is not None:
+            self.stream.synchronize()
+        else:
+            stream_module.get_current_stream().synchronize()
+        assert b_cpu.value == a_cpu.value
+
+
 # -----------------------------------------------------------------------------
 # Memory pool
 
@@ -162,7 +269,7 @@ class TestSingleDeviceMemoryPool(unittest.TestCase):
         self.pool = memory.SingleDeviceMemoryPool(allocator=mock_alloc)
         self.unit = memory._allocation_unit_size
         self.stream = stream_module.Stream()
-        self.stream_ptr = self.stream.ptr
+        self.stream_ident = self.stream.ptr
 
     def test_round_size(self):
         assert memory._round_size(self.unit - 1) == self.unit
@@ -176,20 +283,20 @@ class TestSingleDeviceMemoryPool(unittest.TestCase):
 
     def test_split(self):
         mem = MockMemory(self.unit * 4)
-        chunk = memory._Chunk(mem, 0, mem.size, self.stream_ptr)
+        chunk = memory._Chunk(mem, 0, mem.size, self.stream_ident)
         tail = chunk.split(self.unit * 2)
         assert chunk.ptr() == mem.ptr
         assert chunk.offset == 0
         assert chunk.size == self.unit * 2
         assert chunk.prev is None
         assert chunk.next.ptr() == tail.ptr()
-        assert chunk.stream_ptr == self.stream_ptr
+        assert chunk.stream_ident == self.stream_ident
         assert tail.ptr() == mem.ptr + self.unit * 2
         assert tail.offset == self.unit * 2
         assert tail.size == self.unit * 2
         assert tail.prev.ptr() == chunk.ptr()
         assert tail.next is None
-        assert tail.stream_ptr == self.stream_ptr
+        assert tail.stream_ident == self.stream_ident
 
         tail_of_head = chunk.split(self.unit)
         assert chunk.ptr() == mem.ptr
@@ -197,13 +304,13 @@ class TestSingleDeviceMemoryPool(unittest.TestCase):
         assert chunk.size == self.unit
         assert chunk.prev is None
         assert chunk.next.ptr() == tail_of_head.ptr()
-        assert chunk.stream_ptr == self.stream_ptr
+        assert chunk.stream_ident == self.stream_ident
         assert tail_of_head.ptr() == mem.ptr + self.unit
         assert tail_of_head.offset == self.unit
         assert tail_of_head.size == self.unit
         assert tail_of_head.prev.ptr() == chunk.ptr()
         assert tail_of_head.next.ptr() == tail.ptr()
-        assert tail_of_head.stream_ptr == self.stream_ptr
+        assert tail_of_head.stream_ident == self.stream_ident
 
         tail_of_tail = tail.split(self.unit)
         assert tail.ptr() == chunk.ptr() + self.unit * 2
@@ -211,17 +318,17 @@ class TestSingleDeviceMemoryPool(unittest.TestCase):
         assert tail.size == self.unit
         assert tail.prev.ptr() == tail_of_head.ptr()
         assert tail.next.ptr() == tail_of_tail.ptr()
-        assert tail.stream_ptr == self.stream_ptr
+        assert tail.stream_ident == self.stream_ident
         assert tail_of_tail.ptr() == mem.ptr + self.unit * 3
         assert tail_of_tail.offset == self.unit * 3
         assert tail_of_tail.size == self.unit
         assert tail_of_tail.prev.ptr() == tail.ptr()
         assert tail_of_tail.next is None
-        assert tail_of_tail.stream_ptr == self.stream_ptr
+        assert tail_of_tail.stream_ident == self.stream_ident
 
     def test_merge(self):
         mem = MockMemory(self.unit * 4)
-        chunk = memory._Chunk(mem, 0, mem.size, self.stream_ptr)
+        chunk = memory._Chunk(mem, 0, mem.size, self.stream_ident)
         chunk_ptr = chunk.ptr()
         chunk_offset = chunk.offset
         chunk_size = chunk.size
@@ -244,7 +351,7 @@ class TestSingleDeviceMemoryPool(unittest.TestCase):
         assert head.size == head_size
         assert head.prev is None
         assert head.next.ptr() == tail_ptr
-        assert head.stream_ptr == self.stream_ptr
+        assert head.stream_ident == self.stream_ident
 
         tail.merge(tail_of_tail)
         assert tail.ptr() == tail_ptr
@@ -252,7 +359,7 @@ class TestSingleDeviceMemoryPool(unittest.TestCase):
         assert tail.size == tail_size
         assert tail.prev.ptr() == head_ptr
         assert tail.next is None
-        assert tail.stream_ptr == self.stream_ptr
+        assert tail.stream_ident == self.stream_ident
 
         head.merge(tail)
         assert head.ptr() == chunk_ptr
@@ -260,7 +367,7 @@ class TestSingleDeviceMemoryPool(unittest.TestCase):
         assert head.size == chunk_size
         assert head.prev is None
         assert head.next is None
-        assert head.stream_ptr == self.stream_ptr
+        assert head.stream_ident == self.stream_ident
 
     def test_alloc(self):
         p1 = self.pool.malloc(self.unit * 4)
@@ -509,7 +616,7 @@ class TestSingleDeviceMemoryPool(unittest.TestCase):
             self.pool.set_limit(fraction=1.1)
 
     def test_parse_limit_string(self):
-        parse_limit_string = self.pool._parse_limit_string
+        parse_limit_string = memory._parse_limit_string
 
         # size
         param = parse_limit_string('0')
@@ -545,25 +652,28 @@ class TestSingleDeviceMemoryPool(unittest.TestCase):
 class TestMemoryPool(unittest.TestCase):
 
     def setUp(self):
-        self.pool = memory.MemoryPool(self.allocator)
         if (cupy.cuda.runtime.is_hip
                 and self.allocator is memory.malloc_managed):
             raise unittest.SkipTest('HIP does not support managed memory')
+        self.pool = memory.MemoryPool(self.allocator)
+
+    def tearDown(self):
+        self.pool.free_all_blocks()
 
     def test_zero_size_alloc(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             mem = self.pool.malloc(0).mem
             assert isinstance(mem, memory.Memory)
             assert not isinstance(mem, memory.PooledMemory)
 
     def test_double_free(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             mem = self.pool.malloc(1).mem
             mem.free()
             mem.free()
 
     def test_free_all_blocks(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             mem = self.pool.malloc(1).mem
             assert isinstance(mem, memory.BaseMemory)
             assert isinstance(mem, memory.PooledMemory)
@@ -574,13 +684,13 @@ class TestMemoryPool(unittest.TestCase):
             assert self.pool.n_free_blocks() == 0
 
     def test_free_all_blocks_without_malloc(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             # call directly without malloc.
             self.pool.free_all_blocks()
             assert self.pool.n_free_blocks() == 0
 
     def test_free_all_free(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             mem = self.pool.malloc(1).mem
             assert isinstance(mem, memory.BaseMemory)
             assert isinstance(mem, memory.PooledMemory)
@@ -592,27 +702,27 @@ class TestMemoryPool(unittest.TestCase):
             assert self.pool.n_free_blocks() == 0
 
     def test_free_all_free_without_malloc(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             # call directly without malloc.
             with testing.assert_warns(DeprecationWarning):
                 self.pool.free_all_free()
             assert self.pool.n_free_blocks() == 0
 
     def test_n_free_blocks_without_malloc(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             # call directly without malloc/free_all_free.
             assert self.pool.n_free_blocks() == 0
 
     def test_used_bytes(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             assert 0 == self.pool.used_bytes()
 
     def test_free_bytes(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             assert 0 == self.pool.free_bytes()
 
     def test_total_bytes(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             assert 0 == self.pool.total_bytes()
 
 
@@ -629,7 +739,7 @@ class TestAllocator(unittest.TestCase):
         memory.set_allocator(self.old_pool.malloc)
 
     def test_set_allocator(self):
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             assert 0 == self.pool.used_bytes()
             arr = cupy.arange(128, dtype=cupy.int64)
             assert 1024 == arr.data.mem.size
@@ -663,6 +773,7 @@ class TestAllocator(unittest.TestCase):
 
     def test_allocator_thread_local(self):
         def thread_body(self):
+            cupy.cuda.Device().use()
             new_pool = memory.MemoryPool()
             with cupy.cuda.using_allocator(new_pool.malloc):
                 assert memory.get_allocator() == new_pool.malloc
@@ -673,7 +784,7 @@ class TestAllocator(unittest.TestCase):
                 threading.Barrier(2)
             assert memory.get_allocator() == self.pool.malloc
 
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             t = threading.Thread(target=thread_body, args=(self,))
             t.daemon = True
             t.start()
@@ -695,24 +806,57 @@ class TestAllocator(unittest.TestCase):
         assert arr.data.mem.size == new_pool.used_bytes()
         assert arr.sum() == 128
 
-    def test_reuse_between_thread(self):
-        def job(self):
-            cupy.arange(16)
+    def _reuse_between_thread(self, stream_main, stream_sub):
+        new_pool = memory.MemoryPool()
+
+        def job(stream):
+            cupy.cuda.Device().use()
+            with cupy.cuda.using_allocator(new_pool.malloc):
+                with stream:
+                    arr = cupy.arange(16)
+            self._ptr = arr.data.ptr
+            del arr
             self._error = False
 
         # Run in main thread.
+        self._ptr = -1
         self._error = True
-        job(self)
+        job(stream_main)
         assert not self._error
+        main_ptr = self._ptr
 
         # Run in sub thread.
+        self._ptr = -1
         self._error = True
-        with cupy.cuda.Device(0):
-            t = threading.Thread(target=job, args=(self,))
+        with cupy.cuda.Device():
+            t = threading.Thread(target=job, args=(stream_sub,))
             t.daemon = True
             t.start()
             t.join()
         assert not self._error
+        return main_ptr, self._ptr
+
+    def test_reuse_between_thread(self):
+        stream = cupy.cuda.Stream.null
+        main_ptr, sub_ptr = self._reuse_between_thread(stream, stream)
+        assert main_ptr == sub_ptr
+
+    def test_reuse_between_thread_same_stream(self):
+        stream = cupy.cuda.Stream()
+        main_ptr, sub_ptr = self._reuse_between_thread(stream, stream)
+        assert main_ptr == sub_ptr
+
+    def test_reuse_between_thread_different_stream(self):
+        stream1 = cupy.cuda.Stream()
+        stream2 = cupy.cuda.Stream()
+        main_ptr, sub_ptr = self._reuse_between_thread(stream1, stream2)
+        assert main_ptr != sub_ptr
+
+    @pytest.mark.skipif(cupy.cuda.runtime.is_hip, reason='No PTDS on HIP')
+    def test_reuse_between_thread_ptds(self):
+        stream = cupy.cuda.Stream.ptds
+        main_ptr, sub_ptr = self._reuse_between_thread(stream, stream)
+        assert main_ptr != sub_ptr
 
 
 @testing.gpu
@@ -726,7 +870,7 @@ class TestAllocatorDisabled(unittest.TestCase):
 
     def _check_pool_not_used(self):
         used_bytes = self.pool.used_bytes()
-        with cupy.cuda.Device(0):
+        with cupy.cuda.Device():
             arr = cupy.arange(128, dtype=cupy.int64)
             assert 0 == self.pool.used_bytes() - used_bytes
             del arr
@@ -808,3 +952,136 @@ class TestExceptionPicklable(unittest.TestCase):
         e2 = pickle.loads(pickle.dumps(e1))
         assert e1.args == e2.args
         assert str(e1) == str(e2)
+
+
+@testing.gpu
+@pytest.mark.skipif(cupy.cuda.runtime.is_hip,
+                    reason='HIP does not support async allocator')
+@pytest.mark.skipif(cupy.cuda.driver.get_build_version() < 11020,
+                    reason='malloc_async is supported since CUDA 11.2')
+class TestMallocAsync(unittest.TestCase):
+
+    def setUp(self):
+        self.old_pool = cupy.get_default_memory_pool()
+        memory.set_allocator(memory.malloc_async)
+
+    def tearDown(self):
+        memory.set_allocator(self.old_pool.malloc)
+
+    def _check_pool_not_used(self):
+        used_bytes = self.old_pool.used_bytes()
+        with cupy.cuda.Device():
+            arr = cupy.arange(128, dtype=cupy.int64)
+            assert 0 == self.old_pool.used_bytes() - used_bytes
+            del arr
+
+    def test(self):
+        self._check_pool_not_used()
+
+    def test_stream1(self):
+        # Check: pool is not used when on a stream
+        s = cupy.cuda.Stream()
+        with s:
+            self._check_pool_not_used()
+
+    def test_stream2(self):
+        # Check: the memory was allocated on the right stream
+        s = cupy.cuda.Stream()
+        with s:
+            memptr = memory.alloc(100)
+            assert memptr.mem.stream == s.ptr
+
+    def test_stream3(self):
+        # Check: destory stream does not affect memory deallocation
+        s = cupy.cuda.Stream()
+        with s:
+            memptr = memory.alloc(100)
+
+        del s
+        gc.collect()
+        del memptr
+
+    def test_stream4(self):
+        # Check: free on the same stream
+        s = cupy.cuda.Stream()
+        with s:
+            memptr = memory.alloc(100)
+            del memptr
+
+    def test_stream5(self):
+        # Check: free on another stream
+        s1 = cupy.cuda.Stream()
+        with s1:
+            memptr = memory.alloc(100)
+        del s1
+
+        s2 = cupy.cuda.Stream()
+        with s2:
+            del memptr
+
+
+@testing.gpu
+@pytest.mark.skipif(cupy.cuda.runtime.is_hip,
+                    reason='HIP does not support async allocator')
+@pytest.mark.skipif(cupy.cuda.driver.get_build_version() < 11020,
+                    reason='malloc_async is supported since CUDA 11.2')
+class TestMemoryAsyncPool(unittest.TestCase):
+
+    def setUp(self):
+        self.pool = memory.MemoryAsyncPool()
+        cupy.get_default_memory_pool().free_all_blocks()
+        cupy.cuda.Device().synchronize()
+
+    def test_zero_size_alloc(self):
+        with cupy.cuda.Device():
+            mem = self.pool.malloc(0).mem
+            assert isinstance(mem, memory.MemoryAsync)
+            assert not isinstance(mem, memory.PooledMemory)
+
+    def test_alloc(self):
+        with cupy.cuda.Device():
+            mem = self.pool.malloc(100).mem
+            assert isinstance(mem, memory.MemoryAsync)
+            assert not isinstance(mem, memory.PooledMemory)
+
+    @testing.slow
+    def test_alloc_large_chunk(self):
+        self.pool.free_all_blocks()
+        with cupy.cuda.Device() as d:
+            _, mem_total = d.mem_info
+            mem = self.pool.malloc(int(0.7 * mem_total)).mem  # 70% memory
+            del mem
+            mem = self.pool.malloc(int(0.3 * mem_total)).mem  # 30% memory # noqa
+
+    def test_free_all_blocks(self):
+        with cupy.cuda.Device():
+            mem = self.pool.malloc(1).mem
+            del mem
+            self.pool.free_all_blocks()
+
+    @testing.slow
+    def test_free_all_blocks_large_chunk(self):
+        # When memory is returned to the async mempool, it is not immediately
+        # visible to normal malloc routines until after a sync happens.
+        default_pool = cupy.get_default_memory_pool()
+        with cupy.cuda.Device() as d:
+            _, mem_total = d.mem_info
+            mem = self.pool.malloc(int(0.7 * mem_total)).mem  # 70% memory
+            del mem
+            with pytest.raises(memory.OutOfMemoryError):
+                default_pool.malloc(int(0.3 * mem_total))  # 30% memory
+            self.pool.free_all_blocks()  # synchronize
+            default_pool.malloc(int(0.3 * mem_total))  # this time it'd work
+
+    @testing.slow
+    def test_interaction_with_CuPy_default_pool(self):
+        # Test saneness of cudaMallocAsync
+        default_pool = cupy.get_default_memory_pool()
+        with cupy.cuda.Device() as d:
+            _, mem_total = d.mem_info
+            mem = default_pool.malloc(int(0.7 * mem_total)).mem  # 70% memory
+            del mem
+            with pytest.raises(memory.OutOfMemoryError):
+                self.pool.malloc(int(0.3 * mem_total))  # 30% memory
+            default_pool.free_all_blocks()
+            self.pool.malloc(int(0.3 * mem_total))  # this time it'd work
